----------------------------------------------------------------------
----                                                              ----
---- MC68000 compatible IP Core					                  ----
----                                                              ----
---- This file is part of the SUSKA ATARI clone project.          ----
---- http://www.experiment-s.de                                   ----
----                                                              ----
---- Description:                                                 ----
---- This model provides an opcode and bus timing compatible ip   ----
---- core compared to Motorola's MC68000 microprocessor.          ----
----                                                              ----
---- This file contains a test module for the MOVEM controller.   ----
---- This module is part of the WF68K00_IP main controller.       ----
----                                                              ----
---- Author(s):                                                   ----
---- - Wolfgang Foerster, wf@experiment-s.de; wf@inventronik.de   ----
----                                                              ----
----------------------------------------------------------------------
----                                                              ----
---- Copyright (C) 2006 Wolfgang Foerster                         ----
----                                                              ----
---- This source file may be used and distributed without         ----
---- restriction provided that this copyright statement is not    ----
---- removed from the file and that any derivative work contains  ----
---- the original copyright notice and the associated disclaimer. ----
----                                                              ----
---- This source file is free software; you can redistribute it   ----
---- and/or modify it under the terms of the GNU Lesser General   ----
---- Public License as published by the Free Software Foundation; ----
---- either version 2.1 of the License, or (at your option) any   ----
---- later version.                                               ----
----                                                              ----
---- This source is distributed in the hope that it will be       ----
---- useful, but WITHOUT ANY WARRANTY; without even the implied   ----
---- warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      ----
---- PURPOSE. See the GNU Lesser General Public License for more  ----
---- details.                                                     ----
----                                                              ----
---- You should have received a copy of the GNU Lesser General    ----
---- Public License along with this source; if not, download it   ----
---- from http://www.gnu.org/licenses/lgpl.html                   ----
----                                                              ----
----------------------------------------------------------------------
-- 
-- Revision History
-- 
-- Revision 2K6B  2006/12/24 WF
-- Initial Release.
-- 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity MOVEM_CTRL is
	port (
		CLK				: in bit; -- System clock.
		RESETn			: in bit; -- Core reset.

		ADR_MODE		: in std_logic_vector(2 downto 0);
		BUS_CYC_RDY		: in bit;
		REGLISTMASK		: in bit_vector(15 downto 0);
		STATE			: in bit_vector(1 downto 0);

		MOVEM_EN		: out bit;
		MOVEM_CPY		: out bit;				
		MOVEM_REGSEL	: out std_logic_vector(2 downto 0);
		MOVEM_ADn		: out bit;
		MOVEM_EN_1		: out bit;
		MOVEM_CPY_1		: out bit;				
		MOVEM_REGSEL_1	: out std_logic_vector(2 downto 0);
		MOVEM_ADn_1		: out bit
	);
end entity MOVEM_CTRL;

architecture BEHAVIOR of MOVEM_CTRL is
type OPTYPE is (MOVEM, DUMMY);
type EXEC_STATE_TYPE is (RD_SRC_BASE_ADR_HI, MOVEM_TST, WAITSTATE);
signal OP			: OPTYPE;
signal EXEC_STATE	: EXEC_STATE_TYPE;
signal BIT_PNT_1	: integer range 0 to 15; -- Bit pointer.
begin
	OP <= MOVEM;
	with STATE select
		EXEC_STATE <= 	RD_SRC_BASE_ADR_HI when "00",
						MOVEM_TST when "10",
						WAITSTATE when others;

	MOVEM_CTRL: process(RESETn, CLK)
	-- The MOVEM command takes the REGLISTMASK flags to control, which registers
	-- have to be written or not. The behavior of the reading or writing depends
	-- on the MOVEM operation mode and is controlled in this process.
	variable BIT_PNT	: integer range 0 to 15; -- Bit pointer.
	variable REGSEL_TMP	: std_logic_vector(2 downto 0);
	begin
		if RESETn = '0' then
			BIT_PNT 	:= 0;
			REGSEL_TMP	:= "000";
		elsif CLK = '1' and CLK' event then
			-- Be aware, that the bit pointer starts always at a value of zero. It points 
			-- to the register list mask, which's entries depend on the MOVEM addressing mode.
			if OP = MOVEM and EXEC_STATE = RD_SRC_BASE_ADR_HI and BUS_CYC_RDY = '1' then
				BIT_PNT := 0;
				case ADR_MODE is
					when "100" => REGSEL_TMP := "111"; -- Predecrement mode.
					when others => REGSEL_TMP := "000";
				end case;
			elsif EXEC_STATE = MOVEM_TST and BIT_PNT < 15 then
				BIT_PNT := BIT_PNT + 1;
				case ADR_MODE is
					when "100" => REGSEL_TMP := REGSEL_TMP - '1'; -- Predecrement mode.
					when others => REGSEL_TMP := REGSEL_TMP + '1';
				end case;
			end if;
			--
			case BIT_PNT is
				when 15 => MOVEM_EN <= '0';
				when others => MOVEM_EN <= '1';
			end case;
			--
			MOVEM_CPY <= REGLISTMASK(BIT_PNT);
			--
			if BIT_PNT <= 7 and ADR_MODE = "100" then 
				MOVEM_ADn <= '1'; -- Predecrement address mode.
			elsif BIT_PNT > 7 and ADR_MODE /= "100" then
				MOVEM_ADn <= '1'; -- Other addressing modes.
			else
				MOVEM_ADn <= '0'; -- Select data registers.
			end if;
			--
			MOVEM_REGSEL <= REGSEL_TMP;
		end if;
	end process MOVEM_CTRL;

	MOVEM_CTRL_1: process(RESETn, CLK)
	-- The MOVEM command takes the REGLISTMASK flags to control, which registers
	-- have to be written or not. The behavior of the reading or writing depends
	-- on the MOVEM operation mode and is controlled in this process.
	variable REGSEL_TMP	: std_logic_vector(2 downto 0);
	begin
		if RESETn = '0' then
			BIT_PNT_1 	<= 0;
			REGSEL_TMP	:= "000";
		elsif CLK = '1' and CLK' event then
			-- Be aware, that the bit pointer starts always at a value of zero. It points 
			-- to the register list mask, which's entries depend on the MOVEM addressing mode.
			if OP = MOVEM and EXEC_STATE = RD_SRC_BASE_ADR_HI and BUS_CYC_RDY = '1' then
				BIT_PNT_1 <= 0;
				case ADR_MODE is
					when "100" => REGSEL_TMP := "111"; -- Predecrement mode.
					when others => REGSEL_TMP := "000";
				end case;
			elsif EXEC_STATE = MOVEM_TST and BIT_PNT_1 < 15 then
				BIT_PNT_1 <= BIT_PNT_1 + 1;
				case ADR_MODE is
					when "100" => REGSEL_TMP := REGSEL_TMP - '1'; -- Predecrement mode.
					when others => REGSEL_TMP := REGSEL_TMP + '1';
				end case;
			end if;
			--
			case BIT_PNT_1 is
				when 15 => MOVEM_EN_1 <= '0';
				when others => MOVEM_EN_1 <= '1';
			end case;
			--
			MOVEM_CPY_1 <= REGLISTMASK(BIT_PNT_1);
			--
			if BIT_PNT_1 <= 7 and ADR_MODE = "100" then 
				MOVEM_ADn_1 <= '1'; -- Predecrement address mode.
			elsif BIT_PNT_1 > 7 and ADR_MODE /= "100" then
				MOVEM_ADn_1 <= '1'; -- Other addressing modes.
			else
				MOVEM_ADn_1 <= '0'; -- Select data registers.
			end if;
			--
			MOVEM_REGSEL_1 <= REGSEL_TMP;
		end if;
	end process MOVEM_CTRL_1;
end architecture BEHAVIOR;